use strict;

sub dump_buffer {
  my ($buff, $start, $length) = @_;

  my $end = defined $length ? $length : length($buff);
  $start = 0 if !defined $start;

  my $result = '';
  for (my $i = 0; $i < $end; $i += 16) {
    my $string = substr($buff, $i, 16);
    my $len = length($string);

    $result .= sprintf "%6lx  ", $i + $start;
    $result .= join " ", unpack("H2" x $len, $string);
    substr($result, -3 * ($len - 8), 1) = '-' if $len > 8;
    $result .=  "   " x (16 - $len + 1);

    $string =~ tr/\x20-\x7e/./c;
    $result .= $string;
    $result .= "\n";
  }

  $result;
}

sub read_file {
  my ($file, $size) = @_;

  $size = -s $file;
  open my $fh, '<', $file or die "open: $file\n";
  binmode $fh;
  my $buff;
  read $fh, $buff, $size;
  close $fh;
  return $buff;
}

sub parse_obx {
  my ($buff) = @_;
  
  my $offset = 0;
  my $out = "\xff" x 65536;

  my $length = length($buff);
  while (1) {
    printf STDERR "OFFSET: $offset\n";
    last if $offset + 2 >= $length;
    my $sig = substr($buff, $offset, 2);
    $offset += 2 if $sig eq "\xff\xff";

    last if $offset + 4 >= $length;
    my ($start, $end) = unpack "vv", substr($buff, $offset, 4);
    $offset += 4;

    my $bytes = $end - $start + 1;
    printf STDERR "%04lx: %04lx\n", $start, $bytes;

    substr($out, $start, $bytes) = substr($buff, $offset, $bytes);
    $offset += $bytes;
  }

  $out;
}

sub main {
  my $buff = read_file(shift);
  my $xex = parse_obx($buff);
  print dump_buffer($xex);
}

main(@ARGV);
