use strict;

my $proc = 0;
my $line = 0;
my $next = 0;
my $out = '';

while (<>) {
  $line++;
  chomp;

  print STDERR "CHAR: $line: $_\n" if /[^\x20-\x3fA-Z\[\]|@\\_\^]/; # detect odd characters
  print STDERR "SPACE: $line: $_\n" if /\s+$/;
  print STDERR "COMMENT+: $line: $_\n" if /^.{51};/;
  print STDERR "COMMENT-: $line: $_\n" if /^.{49};/;

  my ($output, $rest) = (/^(.{0,18})(.*)/);
  die "$line: $_\n" if !defined $output;

  # Validate output prefix
  my ($addr, $vals, $count, $branch, $disabled) = (undef, '', 0, 0, 0);
  if ($output =~ /^([0-9A-F]{4}| {4})  = ([0-9A-F]{4}) {6}$/) {
  }
  elsif ($output =~ /^([0-9A-F]{4})  ((?:[0-9A-F][0-9A-F]){0,5})\s+$/) {
    ($addr, $vals) = ($1, $2);
  }
  elsif ($output =~ /^([0-9A-F]{4})  ([0-9A-F]{4}) \^([0-9A-F]{4})\s+$/) {
    ($addr, $vals, $branch) = ($1, $2, $3);
  }
  elsif ($output =~ /^   \- {14}$/) {
    $disabled = 1;
  }
  elsif ($output =~ /^ {18}$/) {
  }
  elsif ($output eq '') {
  }
  else {
    die "$line: OUTPUT: $output\n";
  }

  if (defined $addr) {
    $addr = hex $addr;
    my @vals = unpack "C*", pack "H*", $vals;
    $count = scalar(@vals);
    if ($count >= 0) {
      print STDERR "ADDR: $line: $_\n" if $addr != $next;
      $out .= $_ for @vals;
    }

    $next = $addr + $count;
  }

  my ($comment) = ($rest =~ /(\s*(?<!');.*)$/);
  $rest =~ s/\s*(?<!');.*$//;
  $rest =~ s/\s+$//;

  # Validate ASM data
  if ($rest =~ /^$/) {
  }
  elsif ($rest =~ /^(\S+)?$/) {
  }
  elsif ($rest =~ /^(:?\w*\s+)= \s*\S+$/) {
  }
  elsif ($rest =~ /^(:?\w*\s+)(DS |DW |DB |SET |EQU |ASSERT )/ && length($1) == 8) {
  }
  elsif ($rest =~ /^\s{8}(ORG |IF |ENDIF|E?PROC|END )/) {
  }
  elsif ($rest =~ /^(:?\w*\s+)(ASL|LSR|RO[LR])     A$/ && length($1) == 8 && $count == 1) {
  }
  elsif ($rest =~ /^(:?\w*\s+)(TX[AS]|TA[XY]|TSX|TYA|IN[XY]|DE[XY])$/ && length($1) == 8 && ($count == 1 || $disabled)) {
  }
  elsif ($rest =~ /^(:?\w*\s+)(BRK|CL[CDIV]|SE[CDI]|RT[IS]|P[HL][AP]|NOP)$/ && length($1) == 8 && ($count == 1 || $disabled)) {
  }
  elsif ($rest =~ /^(:?\w*\s+)(JSR|JMP)     \S+$/ && length($1) == 8 && ($count == 3 || $disabled)) {
  }
  elsif ($rest =~ /^(:?\w*\s+)(LD[AXY]|BIT|CMP|CP[XY])     (\S+$|#)/ && length($1) == 8 && ($count == 2 || $count == 3 || $disabled)) {
  }
  elsif ($rest =~ /^(:?\w*\s+)(ST[AXY])     \S+$/ && length($1) == 8 && ($count == 2 || $count == 3 || $disabled)) {
  }
  elsif ($rest =~ /^(:?\w*\s+)(ADC|SBC|INC|DEC|ASL|LSR|RO[LR]|AND|EOR|ORA)     \S+$/ && length($1) == 8 && ($count == 2 || $count == 3 || $disabled)) {
  }
  elsif ($rest =~ /^(:?\w*\s+)(BNE|BEQ|BC[CS]|BMI|BPL|BV[CS])     \S+$/ && length($1) == 8 && (($branch && $count == 2) || $disabled)) {
    if (!$disabled) {
      # Validate branch target
      my $offset = hex substr($vals, 2, 2);
      $offset = $offset - 0x100 if $offset > 0x7f;
      my $tgt0 = $addr + 2 + $offset;
      my $tgt = hex $branch;
      die "$line: $tgt != $tgt0: $_\n" if $tgt != $tgt0;
    }
  }
  elsif ($rest =~ /^\*/) {
    $comment = $rest;
    $rest = '';
  }
  else {
    die "$line: REST: $rest";
  }

  $proc++ if $rest =~ /\bPROC\b/;
  $rest =~ s/(^|  |\(|#|,|\-):/\1L${proc}_/g;

  $rest =~ s/^(\s+(EPROC|PROC|END)\b)/;\1/;
  $rest =~ s/\bDB /\.BYTE /;
  $rest =~ s/\bDW /\.WORD /;
  $rest =~ s/\bDS /\.DS /;
  $rest =~ s/\bIF /\.IF /;
  $rest =~ s/\bENDIF\b/\.ENDIF/;
  $rest =~ s/# *LOW /#</;
  $rest =~ s/# *HIGH /#>/;
  $rest =~ s/\b(ASSERT\b.*)/;\1/;
  $rest =~ s/(\s+(?:ASL|LSR|ROR|ROL)\s+)A(\s+|$)/\1\2/;

  $rest =~ s/$/$comment/;

  print "$rest\n";
}
